import logging

import numpy as np
import torch
import torch.nn as nn

log = logging.getLogger(__name__)

def weighted_crossentropyloss(dataset, classes=None):
    """
    torch.nn.CrosEntropyLoss with class weights inferred from given dataset.
    Parameters
    ----------
    dataset : torch.utils.data.Dataset object.
        If the dataset supports .get_y() function or .get_y_all() function,
        all y values are retreived using the methods. If the dataset does not support such functios,
        [data['y'] for data in dataset] is called which may be very slow if loading data['x'] is heavy.
    classes: list of all possible classes, default=None
        If the dataset argument does not have any instances of some class,
        classes argument can be given to ensure propose size of class weights are applied.
        Missing labels are weighted with 0.
    Returns
    -------
    criterion: nn.CrossEntropyLoss object
        nn.CrossEntropyLoss with class weights applied
    """
    # Loss function
    if hasattr(dataset, 'get_y_all'):
        y = np.array(dataset.get_y_all())
    elif hasattr(dataset, 'get_y'):
        y = np.array([dataset.get_y(idx) in idx in range(len(dataset))])
    else:
        y = np.array([data['y'] for data in dataset])
    y_unique, y_count = np.unique(y, return_counts=True)
    y_count = y_count[np.argsort(y_unique)] # Sort
    if classes is None:
        class_weight = torch.tensor(1/y_count, dtype=torch.float)
        class_weight /= class_weight.sum()
    else:
        log.info(f'[weighted_crossentropyloss] classes found, expanding weight vector to: {len(classes)}')
        class_weight = torch.zeros(len(classes))
        assert set(y_unique).issubset(set(classes))
        for c, y_count_ in zip(y_unique, y_count):
            class_weight[c] = 1/y_count_
        class_weight /= class_weight.sum()

    criterion = nn.CrossEntropyLoss(weight = class_weight)
    log.info('weighted_crossentropyloss - CrossEntropyLoss')
    log.info(f'y_count: {y_count}')
    log.info(f'CrossEntropy Weighted: {criterion.weight}')

    return criterion



import os
import logging

from omegaconf import OmegaConf
import torch

import tools as T
import tools.torch

log = logging.getLogger(__name__)

# %%
'''
Should eliminate the following two.
Too inconsistent
'''
def exp_path(path=None, makedirs=True):
    path = '.' if path is None else path
    exp_path = T.Path(path)
    exp_path.MODEL = 'model' # Save model while training
    exp_path.RESULT = 'result' # Training results
    exp_path.NODE = 'node' # Save node info & model. Package to be communicated
    if makedirs:
        exp_path.makedirs()
    return exp_path

def exp_setting(cfg, path=None):
    '''
    cfg must have the following structure:

    cfg:
      gpu_id: 0 # id of gpu, may be None
      random:
        seed: [int] # random seed
        strict: [bool] # strict random. If True, torch.backends.cudnn.benchmark=False and torch.backends.cudnn.deterministic=True
    '''
    if 'gpu_id' in cfg:
        gpu_id = cfg.gpu_id
    else:
        log.info('gpu_id not found in cfg')
        gpu_id = None

    # TODO: elastic adjustment of cfg.random

    # Print current experiment info
    log.info(OmegaConf.to_yaml(cfg))
    log.info(os.getcwd())

    # Set GPU for current experiment
    device = T.torch.multiprocessing_device(gpu_id) # To use distributed gpus, gpu_id must be None!
    log.info(device)

    T.random.seed(cfg.random.seed, strict=cfg.random.strict)
    path = exp_path(path)
    return device, path

# %%
def model_type(model):
    '''
    Return library type of the given model.
    '''
    if issubclass(type(model), sklearn.BaseEstimator):
        return 'sklearn'
    elif issubclass(type(model), torch.nn.Module):
        return 'torch'
    else:
        return 'unknown'

def extract_dataset(dataset):
    if hasattr(dataset, 'get_x_all'):
        x = dataset.get_x_all()


    return {'x': x, 'y': y}

# %%
class Dataset(torch.utils.data.Dataset):
    '''
    inherit this class and update: __init__,
    '''
    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        if self._i < len(self):
            data = self[self._i]
            self._i +=1
            return data
        else:
            raise StopIteration

    def __getitem__(self, idx):
        x, y = self.get_x(idx), self.get_y(idx)

        return {'x': x, 'y': y}

    # def __len__(self):
    #     return len(self.y)

    def get_x(self, idx):
        # Should return x
        pass
    def get_y(self, idx):
        # Should return y
        pass

    def get_y_all(self):
        # Optional for fast computation when getting loss weights. return all y
        pass

class SimpleDataset(torch.utils.data.Dataset):
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def get_x(self, idx):
        return torch.as_tensor(self.x[idx])

    def get_y(self, idx):
        return torch.as_tensor(self.y[idx])

    def get_y_all(self):
        return torch.as_tensor(self.y)

    def __len__(self):
        return len(self.y)

def reproducible_worker_dict():
    '''Generate separate random number generators for workers,
    so that the global random state is not consumed,
    thereby ensuring reproducibility'''
    def seed_worker(worker_id):
        worker_seed = torch.initial_seed() % 2**32
        numpy.random.seed(worker_seed)
        random.seed(worker_seed)

    g = torch.Generator()
    g.manual_seed(0)
    return {'worker_init_fn': seed_worker, 'generator': g}



import logging

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data as D

import tools as T
import tools.torch

# %%
log = logging.getLogger(__name__)

# %%
'''
feature visualization module
'''


"""

Parameters
----------
objective : nn.Module object which takes an input and outputs a single value to optimize.
    Recommended way to make this object is to build another class inheriting nn.Module,
    and constructing the objective function within the class.

See Also
--------
Output_i : nn.Module which computes i-th output of the given model.
    Useful to construct simple objective function.

Examples
--------
>>>
"""

class Output_i(nn.Module):
    """
    Output objective

    Parameters
    ----------
    model : nn.Module object
        Model to optimize.
    class_i: int
        Output class index.
    maximize: bool, default=True
        If True, then forward returns negated value (since loss is minimized in pytorch),
    suppress: bool, default=False
        If True, then forward adds non-negated values of other classes.
        Suppresses other class activations.

    Examples
    --------
    >>> log.info('hello')
    hello
    """
    def __init__(self, model, class_i, maximize=True, suppress=False):
        super().__init__()
        self.model = model
        self.class_i = class_i
        self.maximize = maximize
        self.suppress = suppress

    def forward(self, x):
        y_hat = self.model(x)
        objective = y_hat[:, self.class_i]
        if self.suppress:
            class_i_suppress = list(range(y_hat.shape[-1]))
            class_i_suppress.remove(self.class_i)
            objective = objective - y_hat[:, class_i_suppress].mean(-1)

        if self.maximize:
            objective = -objective
        # if self.maximize:
        #     objective = -y_hat[:, self.class_i]
        # else:
        #     objective = y_hat[:, self.class_i]
        return objective

class RandomSampler():
    '''
    Sample random data
    '''
    allowed_methods = ['rand', 'randn']
    def __init__(self, input_shape, method='rand'):
        self.input_shape = input_shape
        self.method = method
        assert method in self.allowed_methods, f'allowed methods are: {self.allowed_methods}\nreceived: {method}'

    def sample(self, n_sample, method=None):
        method = self.method if method is None else method
        if method == 'rand':
            sample = torch.rand(n_sample, *self.input_shape)
        elif method == 'randn':
            sample = torch.randn(n_sample, *self.input_shape)
        return sample

# %%
def _optimize(objective, sample, updates, Optimizer, optimizer_kwargs, preprocess=None, regularization=None, threshold=None, return_best=False, verbose=False):
    device_original = sample.device
    device = T.torch.get_device(objective)
    sample = sample.clone().detach().to(device) # clone sample so that original tensor will not be modified.
    sample.requires_grad_(True)

    # if not sample.requires_grad:
    #     warnings.warn('Provide sample with sample.requires_grad_(True).\
    #     _optmize() function will automatically setting requires_grad to True, but this may resuilt in undesired results.')

    optimizer_ = Optimizer([sample], **optimizer_kwargs)

    if preprocess is not None:
        log.info('preprocess specified')
    if regularization is not None:
        log.info('regularization specified')

    if verbose:
        with torch.no_grad():
            sample_processed = sample if preprocess is None else preprocess(sample)
            sample_regularized = sample if regularization is None else regularization(sample_processed)
            y_hat = objective(sample_regularized)
            loss = y_hat.mean()
            log.info(f'[iter: 0/{updates}][loss: {loss.item()}]')

    n_verbose = updates // 10

    if threshold is not None:
        threshold_crossed_i = np.zeros(len(sample), dtype=np.bool)
        sample_optimized = torch.empty_like(sample)

    for i in range(1, updates+1):
        sample_processed = sample if preprocess is None else preprocess(sample)
        sample_regularized = sample if regularization is None else regularization(sample_processed)
        y_hat = objective(sample_regularized)
        loss = y_hat.mean()

        # Note this is threshold after preprocessing & regularization(!!)
        if threshold is not None:
            with torch.no_grad():
                # Store sample when threshold passed
                threshold_crossed_i_ = (y_hat.detach().cpu().numpy() < threshold)
                store_i = ~threshold_crossed_i&threshold_crossed_i_
                sample_optimized[store_i] = sample[store_i].detach()
                threshold_crossed_i[store_i] = True

                if store_i.any():
                    log.info('threshold!')
                    log.info(store_i, sample_optimized[store_i].detach(), sample_regularized[store_i].detach(), y_hat)
                # # if threshold_crossed_i_.any():
                # #     log.info('threshold!')
                # #     break

                if threshold_crossed_i_.all():
                    log.info(f'Threshold crossed. [-loss: ({-loss.item()}) > threshold: ({threshold})]')
                    break

        optimizer_.zero_grad()
        loss.backward()
        optimizer_.step()
        if i%n_verbose==0:
            log.info(f'[iter: {i}/{updates}][loss: {loss.item()}]')

    # Should return preprocessed sample?
    if threshold is not None:
        sample_optimized[~threshold_crossed_i] = sample[~threshold_crossed_i].detach()
        return sample_optimized.to(device_original)
    else:
        return sample.detach().to(device_original)

def optimize(objective, sample, updates, Optimizer, optimizer_kwargs, preprocess=None, regularization=None, threshold=None, return_best=False, verbose=False, batch_size=None):
    """
    Optimize sample
    Copies the sample and returns the optimized tensor.

    Parameters
    ----------
    arg :
        dd
    Returns
    -------
    sample : torch.Tensor which is optimized

    References
    ----------
    .. [1] `Feature Visualization
           <https://distill.pub/2017/feature-visualization/>`_
    Examples
    --------
    >>>
    """
    if batch_size is None:
        sample_optimized = _optimize(objective=objective, sample=sample, updates=updates, preprocess=preprocess, regularization=regularization, Optimizer=Optimizer, optimizer_kwargs=optimizer_kwargs, threshold=threshold, return_best=return_best, verbose=verbose)
    else:
        if verbose: log.info('Run in batches')
        loader = D.DataLoader(sample, batch_size=batch_size, drop_last=False)
        l_sample_optimized = []
        for sample_batch in loader:
            sample_optimized = _optimize(objective=objective, sample=sample, updates=updates, preprocess=preprocess, regularization=regularization, Optimizer=Optimizer, optimizer_kwargs=optimizer_kwargs, threshold=threshold, return_best=return_best, verbose=verbose)
            l_sample_optimized.append(sample_optimized)
        sample_optimized = torch.stack(l_sample_optimized, axis=0)
    return sample_optimized



from copy import deepcopy as dcopy
import logging
import multiprocessing

import numpy as np
import sklearn.metrics as metrics
import torch

import tools as T
import tools.torch
import tools.numpy
import torch.utils.data as D

# %%
log = logging.getLogger(__name__)

# %%
'''
Testing functions

give x and produce y_hat, y_score, y_pred, etc...
'''
def transfer_misc(node, dataset, verbose=False):
    misc_temp = T.TDict()
    transfer_log = []

    if 'get_f' in node.misc:
        transfer_log.append('get_f')
        misc_temp.get_f, dataset.get_f = dataset.get_f, node.misc.get_f

    if len(transfer_log)!=0:
        log.info(f'Transferring misc from node ({node.name}) -> dataset: {transfer_log}')

    return dataset, misc_temp

def inverse_transfer_misc(misc_temp, dataset):
    transfer_log = []
    if 'get_f' in misc_temp:
        transfer_log.append('get_f')
        dataset.get_f = misc_temp.get_f

    if len(transfer_log)!=0:
        log.info(f'Inverse transferring misc -> dataset: {transfer_log}')

    return dataset

def reproducible_worker_dict():
    '''Generate separate random number generators for workers,
    so that the global random state is not consumed,
    thereby ensuring reproducibility'''
    import random
    def seed_worker(worker_id):
        worker_seed = torch.initial_seed() % 2**32
        np.random.seed(worker_seed)
        random.seed(worker_seed)

    g = torch.Generator()
    g.manual_seed(0)
    return {'worker_init_fn': seed_worker, 'generator': g}

targets_type_list = [None, 'categorical', 'multihead_classification', 'autoencode'] # move this to utils?
def _test_assertion(dataset, targets_type, test_f, result_dict, keep_x):
    # Safety check
    if test_f is not None and result_dict is not None:
        assert callable(test_f) and issubclass(type(result_dict), dict), f'test_f must be callable and result_dict must be dict-like'
    elif test_f is None and result_dict is None:
        # Infer targets_type and create corresponding test_f and result_dict
        if targets_type is None:
            if hasattr(dataset, 'targets_type'):
                targets_type = dataset.targets_type
            else:
                raise Exception('When test_f and result_dict is not given, targets_type must be given or the dataset must have the attribute "targets_type"')
        assert targets_type in targets_type_list, f'targets_type must be one of {targets_type_list}, received: {targets_type}'
        test_f = get_test_f(targets_type)
        result_dict = get_result_dict(targets_type, keep_x=keep_x)
    else:
        raise Exception(f'test_f and result_dict must either both be provided or None, received [test_f: {test_f}][result_dict: {result_dict}]')
    return test_f, result_dict

def test(model, dataset, batch_size=64, targets_type=None, test_f=None, result_dict=None, keep_x=False, num_workers=0, amp=False):
    '''
    inference stage of nn model.
    '''
    test_f, result_dict = _test_assertion(dataset=dataset, targets_type=targets_type, test_f=test_f, result_dict=result_dict, keep_x=keep_x)

    device = T.torch.get_device(model) # Test on the model's device
    model.eval()

    # dataset, misc_temp = transfer_misc(node, dataset) # Assume the dataset is consistent
    kwargs_dataloader = reproducible_worker_dict()
    test_loader = D.DataLoader(dataset, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=num_workers, **kwargs_dataloader)

    with torch.no_grad():
        # Mixed precision for acceleration
        if amp:
            with torch.cuda.amp.autocast():
                for data in test_loader:
                    result_dict = test_f(data=data, model=model, result_dict=result_dict, device=device, keep_x=keep_x)
        else:
            for data in test_loader:
                result_dict = test_f(data=data, model=model, result_dict=result_dict, device=device, keep_x=keep_x)

    result_dict = {k: np.concatenate(v, axis=0) if len(v)>0 else v for k, v in result_dict.items()}

    model.train() # Is this necessary?


    return result_dict

def test_node(node, dataset, batch_size=64, targets_type=None, test_f=None, result_dict=None, keep_x=False, num_workers=0, amp=False):
    '''
    wrapper around node.
    transfers get_f (preprocessing modules) of node to dataset, and returns them to original dataset after inference.
    '''
    model = node.model
    dataset, misc_temp = transfer_misc(node, dataset)
    result_dict = test(model, dataset, batch_size, targets_type, test_f, result_dict, keep_x, num_workers, amp)
    dataset = inverse_transfer_misc(misc_temp, dataset)

    return result_dict

def get_test_f(targets_type):
    assert targets_type in targets_type_list, f'targets_type must be one of {targets_type_list}, received: {targets_type}'
    if targets_type is None:
        return _test_base
    elif targets_type == 'categorical':
        return _test_categorical
    elif targets_type == 'multihead_classification':
        return _test_multihead_categorical
    elif targets_type == 'autoencode':
        return _test_autoencode
    else:
        raise Exception(f'unknown targets_type: {targets_type}')

def get_result_dict(targets_type, keep_x=False):
    assert targets_type in targets_type_list, f'targets_type must be one of {targets_type_list}, received: {targets_type}'
    if targets_type is None:
        result_list = ['y_true', 'y_hat']
        if keep_x: result_list.append('x')
    elif targets_type in ['categorical', 'multihead_classification']:
        result_list = ['y_true', 'y_hat', 'y_score', 'y_pred']
        if keep_x: result_list.append('x')
    elif targets_type == 'autoencode':
        result_list = ['x', 'z', 'x_hat']
    else:
        raise Exception(f'unknown targets_type: {targets_type}')
    return {r:[] for r in result_list}

def _test_base(data, model, result_dict, device=None, keep_x=False):
    # device = device if device is not None else T.torch.get_device(model)
    x = data['x'].to(device)
    y = data['y'].to(device)
    y_hat = model(x)

    result_dict['y_true'].append(y.cpu().numpy())
    result_dict['y_hat'].append(y_hat.cpu().numpy())
    if keep_x: result_dict['x'].append(x.cpu().numpy())
    return result_dict

def _test_categorical(data, model, result_dict, device=None, keep_x=False):
    # Is y_hat necessary? for torch_crossentropyloss? any other methods?
    x = data['x'].to(device)
    y = data['y'].to(device)
    y_hat = model(x)
    y_score = torch.softmax(y_hat, dim=1) # (N, n_classes)

    result_dict['y_true'].append(y.cpu().numpy())
    result_dict['y_hat'].append(y_hat.cpu().numpy())
    result_dict['y_score'].append(y_score.cpu().numpy())
    result_dict['y_pred'].append(y_score.argmax(axis=-1).cpu().numpy())
    if keep_x: result_dict['x'].append(x.cpu().numpy())
    return result_dict

def _test_multihead_categorical(data, model, result_dict, device=None, keep_x=False):
    x = data['x'].to(device)
    y = data['y'].to(device)
    y_hat = model(x)
    y_score = torch.softmax(y_hat, dim=-1) # (N, subtype, n_classes)

    result_dict['y_true'].append(y.cpu().numpy())
    result_dict['y_hat'].append(y_hat.cpu().numpy())
    result_dict['y_score'].append(y_score.cpu().numpy())
    result_dict['y_pred'].append(y_score.argmax(axis=-1).cpu().numpy())

    return result_dict

def _test_autoencode(data, model, result_dict, device=None):
    x = data['x'].to(device)
    z = model.encoder(x)
    x_hat = torch.sigmoid(model.decoder(z))

    x, z, x_hat = x.cpu().numpy(), z.cpu().numpy(), x_hat.cpu().numpy()

    result_dict['x'].append(x)
    result_dict['z'].append(z)
    result_dict['x_hat'].append(x_hat)

    return result_dict

# %%
'''
Scorers

They all receive dict of numpy arrays.
Predefined keys are:
'y_true', 'y_hat', 'y_pred'

Example
-------
result = {'y_true': y_true, 'y_pred': y_pred}
classification_score(result)

scores = {'scorer type': score, ...}


# order of arguments follow sklearn convention: y_true, y_hat/pred

Note
----
pytorch metrics receive (y_hat/pred, y_true), whereas sklearn receive (y_true, y_hat/pred)
'''

# Regresison scorers
def l1_score(result): # alias
    return metrics.mean_absolute_error(result['y_true'], result['y_hat'])

def l2_score(result): # alias
    return metrics.mean_squared_error(result['y_true'], result['y_hat'])

def mse_score(result): # alias
    return metrics.mean_squared_error(result['y_true'], result['y_hat'])

def r2_score(result):
    return metrics.r2_score(result['y_true'], result['y_hat'])

def p_norm_score(result, p=2): # When using this with Validation object, use functools.partial() to fix ``p``
    score = ((np.abs(result['y_hat']-result['y_true'])**p).sum())**(1/p)
    return score

# Classification score
def accuracy_score(result):
    return metrics.accuracy_score(result['y_true'], result['y_pred'])

# def accuracy_score_y_hat(result): # Use this for Validation object, which feeds y_hat to all of its scorers.
#     return metrics.accuracy_score(result['y_true'], result['y_hat'].argmax(axis=1))

# Binary classification
def sensitivity_score(result): # alias
    '''sensitivity == recall == tpr'''
    return metrics.recall_score(result['y_true'], result['y_pred'])

def specificity_score(result):
    (tn, fp), (fn, tp) = metrics.confusion_matrix(result['y_true'], result['y_pred'])
    if (tn+fp)==0:
        warnings.warn('invalid value in specificity_score, setting to 0.0')
        return 0
    return tn / (tn+fp)

# Multiclass classification
def classification_report_full(result, discard_ovr=False):
    '''
    Adds additional metrics to sklearn.classification_report
    '''
    # TODO: add y_score
    y_score = result['y_score'] if 'y_score' in result else None

    y_true, y_pred = result['y_true'], result['y_pred']
    scores = metrics.classification_report(y_true, y_pred, output_dict=True)
    scores_ovr = {k:dcopy(v) for k, v in scores.items() if k.isnumeric()}
    scores_all = {k:dcopy(v) for k, v in scores.items() if not k.isnumeric()}
    more_scorers = {'sensitivity': sensitivity_score, 'specificity': specificity_score, 'accuracy': accuracy_score} # Optimize to reduce redundant computations?

    # Additional metrics
    for c in scores_ovr.keys():
        c_int = int(c)
        y_true__c, y_pred_c = y_true==c_int, y_pred==c_int
        for scorer_name, scorer in more_scorers.items():
            scores_ovr[c][scorer_name] = scorer({'y_true': y_true__c, 'y_pred': y_pred_c})

    if y_score is not None:
        for c in scores_ovr.keys():
            c_int = int(c)
            y_score_ = y_score[:, c_int]
            metrics.roc_auc_score(y_true, y_score_)

    # summary
    for scorer_name, scorer in more_scorers.items():
        scores_all['macro avg'][scorer_name] = np.mean([scores_ovr_[scorer_name] for scores_ovr_ in scores_ovr.values()])
        scores_all['weighted avg'][scorer_name] = np.sum([scores_ovr_[scorer_name]*scores_ovr_['support'] for scores_ovr_ in scores_ovr.values()]) / scores_all['weighted avg']['support']

    if discard_ovr:
        return scores_all
    else:
        scores_ovr.update(scores_all)
        return scores_ovr

# Multihead classification score
def multihead_accuracy_score(y_pred, y):
    '''
    :param y_pred: array of shape (N, subtype)
    :param y: array of shape (N, subtype)
    '''
    assert y_pred.ndim==2 and y.ndim==2
    score = [accuracy_score(y_pred_, y_) for y_pred_, y_ in zip(y_pred.T, y.T)]
    score = np.mean(score)
    return score

def multihead_accuracy_score_score(y_score, y):
    '''
    :param y_pred: array of shape (N, subtype, n_classes)
    :param y: array of shape (N, subtype)
    '''
    assert y_score.ndim==3 and y.ndim==2
    y_pred = y_score.argmax(axis=-1)
    return multihead_accuracy_score(y_pred, y)

# %%
# All scores for each task type
def regression_score(result):
    '''
    :param result: dict with the following keys: [y_hat, y]
    '''
    y_hat, y = result['y_hat'], result['y']
    l1 = l1_score(y_hat, y)
    mse = mse_score(y_hat, y)
    r2 = r2_score(y_hat, y)

    scores = {
    'l1': l1,
    'mse': mse,
    'r2': r2
    }

    return scores

def classification_score(result, multi_class='ovr', discard_ovr=False):
    '''
    :param result: dict with the following keys: [y_pred, y_score, y]


    y_score: array of shape (N, n_classes)
    y: array of shape (N,)
    '''
    y_pred, y_score, y_true = result['y_pred'], result['y_score'], result['y']

    # classification_result = classification_report_full(y_true, y_pred, discard_ovr=discard_ovr)
    classification_result = classification_report_full(result, discard_ovr=discard_ovr)
    auroc = metrics.roc_auc_score(y_true, y_score, multi_class=multi_class)
    kappa = metrics.cohen_kappa_score(y_true, y_pred)
    m_cc = metrics.matthews_corrcoef(y_true, y_pred)
    c_matrix = metrics.confusion_matrix(y_true, y_pred)

    scores = {
    'auroc': auroc,
    'kappa': kappa,
    'm_cc': m_cc,
    'c_matrix': c_matrix,
    }
    classification_result.update(scores)
    return scores

def multihead_classification_score(result):
    '''
    :param result: dict with the following keys: [y_pred, y_score, y]


    y_score: array of shape (N, subtype, n_classes)
    y: array of shape (N, subtype)
    '''
    y_pred, y_score, y = result['y_pred'], result['y_score'], result['y']
    multihead_accuracy = multihead_accuracy_score(y_pred, y)

    scores = {
    'multihead_accuracy_score': multihead_accuracy
    }
    return scores


# Wrapper function for hydra.utils.instantiate
class Score_wrapper:
    '''
    wraps around scorers to:
    1. instantiate an object using hydra.utils.instantiate()
    2. adjust scorer functions such that raw input can be delivered to scorer functions (y_pred or y_score can be delivered to any type of scorers)
    '''
    f_dict = {
    # regression
    'l1_score': l1_score,
    'mse_score': mse_score,
    'r2_score': r2_score,

    # multihead classification
    'multihead_accuracy_score': multihead_accuracy_score_score,
    }
    def __init__(self, func):
        self.func = self.f_dict[func]

    def __call__(self, y_hat, y):
        return self.func(y_hat, y)
# %%
'''
May be deprecated
'''
def save_score():
    f

def _pr_score(y_true, y_score, threshold):
    y_pred = T.numpy.binarize(y_score, threshold)
    pr = metrics.precision_score(y_true, y_pred)
    rec = metrics.recall_score(y_true, y_pred)
    return pr, rec

def precision_recall_curve_all(y_true, y_score, num_workers=None):
    '''
    compute precision-recall curve for all threshold.
    Because sklearn.metrics.precision_recall_curve does not return full set of thresholds
    sklearn drops the results after when recall hits 1.

    -> is this function really necessary?
    -> may be deprecated
    '''
    thresholds = np.sort(np.unique(y_score)) # increasing threshold
    thresholds = np.append(thresholds,1) # Since tnp.binarize binarizes with y_pred[y_score>=threshold] == 1
    if num_workers==0:
        prs, recs = [], []
        for t in thresholds:
            y_pred = T.numpy.binarize(y_score, threshold)
            prs.append(metrics.precision_score(y_true, y_pred))
            recs.append(metrics.recall_score(y_true, y_pred))
    elif num_workers is None or T.isint(num_workers):
        with multiprocessing.Pool() as p:
            l_pr = p.starmap(_pr_score, zip(it.repeat(y_true, len(thresholds)), it.repeat(y_score, len(thresholds)), thresholds))

    return prs, recs, thresholds
